/* throw.c   by Michael Thorpe   2005-08-13 */

#include "lisp.h"

static obj *access_error;
static obj *argument_error;
static obj *oom_error;
static obj *rangecheck_error;
static obj *type_error;

obj *exception,*exceptionvalue;
obj *exceptionvar,*exceptionvaluevar;

static obj *catch(scope *s,obj *type,obj *func,obj *handler) {
	obj *e,*o,*tmp;

	o=evaluate(incscoperef(s),func);
	if(!o && (type==exception || type==&nilobj)) {
		decref(type);
		s=newscope(s);
		if(!s) {
			decref(handler);
			return(0);
		}
		e=exception;
		exception=incref(&nilobj);
		tmp=define_local(incscoperef(s),incref(exceptionvar),e);
		if(!tmp) {
			decscoperef(s);
			decref(handler);
			return(0);
		}
		decref(tmp);
		e=exceptionvalue;
		exceptionvalue=incref(&nilobj);
		tmp=define_local(incscoperef(s),incref(exceptionvaluevar),e);
		if(!tmp) {
			decscoperef(s);
			decref(handler);
			return(0);
		}
		decref(tmp);
		return(evaluate(s,handler));
	}
	decref(type);
	decscoperef(s);
	decref(handler);
	return(o);
}

obj *throw(obj *type,obj *value) {
	decref(exception);
	decref(exceptionvalue);
	exception=type;
	exceptionvalue=value;
	return(0);
}

obj *throwaccesserror(obj *value) {
	return(throw(incref(access_error),value));
}

obj *throwargumenterror(obj *value) {
	return(throw(incref(argument_error),value));
}

obj *throwoom() {
	return(throw(incref(oom_error),incref(&nilobj)));
}

obj *throwrangecheck(obj *value) {
	return(throw(incref(rangecheck_error),value));
}

obj *throwtypeerror(obj *value) {
	return(throw(incref(type_error),value));
}

int throw_init() {
	if(!(access_error=newword("access-error",12))
	|| !(argument_error=newword("argument-error",14))
	|| !(oom_error=newword("out-of-memory",13))
	|| !(rangecheck_error=newword("rangecheck",10))
	|| !(type_error=newword("type-error",10))
	|| !(exceptionvar=newword("#exception",10))
	|| !(exceptionvaluevar=newword("#exceptionvalue",15))
	|| store_builtin_s3("#catch",catch)
	|| store_builtin_ns2("throw",throw))
		return(1);
	exception=incref(&nilobj);
	exceptionvalue=incref(&nilobj);
	INC_KNOWN_OBJECTS(7);
	return(0);
}

